import json
import jsonref


def format_openapi(openapi_spec: dict) -> str:
    expanded_openapi = jsonref.loads(json.dumps(openapi_spec), proxies=False)
    simplified = remove_keys(
        expanded_openapi, ["responses", "components", "operationId"]
    )
    return json.dumps(simplified)


def remove_keys(obj: dict | list, keys_to_remove: list[str]):
    if isinstance(obj, dict):
        for key in keys_to_remove:
            if key in obj:
                del obj[key]
        for _, value in obj.items():
            remove_keys(value, keys_to_remove)
    elif isinstance(obj, list):
        for item in obj:
            remove_keys(item, keys_to_remove)
    return obj



class QueryParsingException(Exception):
    """Exception raised when a query cannot be parsed into a Rest Request."""

    def __init__(self, message: str):
        self.message = f"""
Unable to parse query into a Rest Request.

Potential reasons:
    - Provided ChatModel is too small 
    - Provided OpenAPI is too vague or large
    
Error message: {message}
"""
        super().__init__(self.message)



import logging

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import PromptTemplate
from requests import Response, request

from rest_ai.exceptions import QueryParsingException
from rest_ai.utilities.config import RETRY_COUNT
from rest_ai.utilities.openapi_utils import format_openapi
from rest_ai.utilities.types import RestRequest, validate_rest_request

from .langchain.prompt_templates import (
    rest_prompt_template,
    rest_prompt_template_with_error,
)


class RestAi:
    base_url: str
    openapi_spec: str
    base_model: BaseChatModel
    chain: any

    def __init__(self, base_url: str, openapi_spec: dict, base_model: BaseChatModel):
        self.base_url = base_url
        self.openapi_spec = format_openapi(openapi_spec)
        self.base_model = base_model
        print(self.openapi_spec)

    def invoke(self, prompt: str) -> Response:
        rest_request = self.extract_request(
            rest_prompt_template, {"query": prompt, "openapi_spec": self.openapi_spec}
        )
        response = self.execute_request(rest_request)
        if response.ok:
            return response
        return self.retry_request_extraction(prompt, response, rest_request)

    def retry_request_extraction(
        self, prompt: str, previous_response: Response, previous_request: RestRequest
    ) -> Response:
        for i in range(RETRY_COUNT):
            logging.info(f"Retry Attempt: {str(i + 1)}")
            rest_request = self.extract_request(
                rest_prompt_template_with_error,
                {
                    "query": prompt,
                    "openapi_spec": self.openapi_spec,
                    "previous_request": previous_request.model_dump_json(),
                    "error": previous_response.text,
                },
            )
            response = self.execute_request(rest_request)
            if response.ok:
                return response
            previous_response = response
            previous_request = rest_request
        logging.error("Unable to successfully invoke query.")

    def extract_request(
        self, prompt_template: PromptTemplate, input: dict
    ) -> RestRequest:
        chain = prompt_template | self.base_model.with_structured_output(RestRequest)
        try:
            query = chain.invoke(input)
            if query is None:
                raise Exception("Parsed query is None.")
        except Exception as e:
            raise QueryParsingException(e)
        logging.info(f"Parsed query: {query}")
        return query

    def execute_request(self, rest_request: RestRequest) -> Response:
        rest_request = validate_rest_request(rest_request)
        response = request(
            method=rest_request.verb,
            url=f"{self.base_url}{rest_request.query_path}",
            params=rest_request.query_params,
            json=rest_request.body,
        )
        logging.info(f"Response: {response.text}")
        return response



from langchain_core.prompts import PromptTemplate

rest_prompt_template = PromptTemplate.from_template(
    """
        Create a valid REST API request based on the user's query. Use the openapi_spec provided to provide the method, query, body and parameters.
        
        query: {query}
        
        openapi_spec: {openapi_spec}
        """
)

rest_prompt_template_with_error = PromptTemplate.from_template(
    """
        Create a valid REST API request based on the user's query. Use the openapi_spec. A previous request was made and it resulted in an error. Use the error message to correct the request.
        
        query: {query}
        
        previous request: {previous_request}
        
        previous_error: {error}
        
        openapi_spec: {openapi_spec}
        """
)



import logging

from pydantic import BaseModel, Field

_valid_verbs = {"GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"}


class RestRequest(BaseModel):
    verb: str = Field(
        description="HTTP verb to use for the request",
        examples=["GET", "POST", "PUT"],
        default="GET",
    )
    query_path: str = Field(
        description="Path to use for the REST request", examples=["/api/v1/users"]
    )
    body: dict | None = Field(
        description="A set of key values to send in the body of the request taken from the openapi spec. Set to None if not applicable",
        default={},
    )
    query_params: dict | None = Field(
        description="A set of key-values to send in the query taken from the openapi spec. Set to None if not applicable.",
        default={},
    )


def validate_rest_request(rest_request: RestRequest) -> RestRequest:
    return RestRequest(
        verb=validate_rest_verb(rest_request.verb),
        query_path=add_preceding_slash(rest_request.query_path),
        body=rest_request.body,
        query_params=rest_request.query_params,
    )


def validate_rest_verb(verb: str) -> str:
    verb = verb.upper().strip()
    if verb not in _valid_verbs:
        logging.warning(f"Invalid verb provided: {verb}. Defaulting to GET.")
        verb = "GET"
    return verb


def add_preceding_slash(path: str) -> str:
    if not path.startswith("/"):
        return "/" + path
    return path
